@@ -4,23 +4,18 @@ Copyright (c) 2021-2022, Wei Wang <wei.wang.bit@outlook.com>
 DearBagPlayer Application
 """
 
-import os
-import sys
-
-script_abspath = os.path.split(os.path.realpath(__file__))[0]
-sys.path.append(script_abspath)
 
 try:
-    import timeline_widgets
+    from .timeline_widgets import TimelineWidgets
 except:
-    raise ModuleNotFoundError('Module timeline_widgets not found.')
+    raise ImportError('Class TimelineWidgets not found.')
 
 import dearpygui.dearpygui as dpg
 import rosbag
 import numpy as np
 import time
 import bisect
-
+import os
 import yaml
 from yaml.loader import SafeLoader
 
@@ -44,7 +39,7 @@ class DearBagPlayer:
         # Timeline
         self.min_time = 0.0
         self.max_time = 0.0
-        self.__timeline = timeline_widgets.TimelineWidgets(start_time=0.0, duration=0.0, loop_enabled=True)
+        self.__timeline = TimelineWidgets(start_time=0.0, duration=0.0, loop_enabled=True)
 
         # Time series control
         self.start_time = 0.0
@@ -56,6 +51,13 @@ class DearBagPlayer:
         # Series selector
         self.tab_bar = None
 
+        # Resize
+        self.viewport_size = [-1, -1]
+        self.scale = [0.3, 0.7]
+        self.plot_size = [-1, -1]
+        self.plot_window_min_width = -1
+        self.data_pool_min_width = -1
+
     def initTimeline(self):
         self.max_time = 0.0
         self.min_time = 1e19
@@ -87,7 +89,11 @@ class DearBagPlayer:
 
             if not checkTopic(topics):
                 bag.close()
-                raise ValueError('At least one given topic not found in rosbag!')
+                self.createErrorPopup(
+                    "At least one given topic not found in rosbag!",
+                    popup_width=350, popup_height=60
+                )
+                return
 
         """
         msg_data = {
@@ -241,6 +247,7 @@ class DearBagPlayer:
     def vlinesTimeUpdate(self, timestamp):
         if not self.vlines:
             self.createTimeLines()
+            return
 
         for vline in self.vlines:
             # TODO: Figure out why built-in float type get errors
@@ -249,6 +256,7 @@ class DearBagPlayer:
     def xypointsUpdate(self):
         if not self.xypoints:
             self.createTimePoints()
+            return
 
         total_index = 0
         last_user_data_length = 0
@@ -281,6 +289,68 @@ class DearBagPlayer:
         if len(dpg.get_item_children(self.tab_bar)[1]) == 1:
             self.addPlotPageCb("Add Plot Button", None, None)
 
+    # -----------------------------------------
+    # Layout
+    # -----------------------------------------
+
+    def resizeCb(self):
+        """
+        Resize data pool window and plot window (including active plot)
+        """
+        self.viewport_size[0] = dpg.get_viewport_width()
+        self.viewport_size[1] = dpg.get_viewport_height()
+        dpg.configure_item('data_pool_window',
+                           width=int(self.scale[0]*self.viewport_size[0]),
+                           height=int(self.viewport_size[1] - 45),
+                           )
+        dpg.configure_item('plot_window',
+                           width=int(self.scale[1]*self.viewport_size[0] - 10),
+                           height=int(self.viewport_size[1] - 45),
+                           )
+        width, height = dpg.get_item_rect_size('plot_window')
+        self.plot_size[0] = width - 45
+        self.plot_size[1] = height - 150
+        plot_tag = dpg.get_item_user_data(self.tab_bar)['act_plot']
+        dpg.configure_item(plot_tag, width=self.plot_size[0], height=self.plot_size[1])
+
+    def addVerticalSeparator(self):
+        separator = dpg.add_button(width=3, height=-1)
+
+        def clickedCb():
+            while dpg.is_mouse_button_down(0):
+                # Calculate new widths
+                x_pos = dpg.get_mouse_pos(local=False)[0]
+                dpg.split_frame(delay=10)
+                x_delta = dpg.get_mouse_pos(local=False)[0] - x_pos
+                width_left = dpg.get_item_width("data_pool_window") + x_delta
+                width_right = dpg.get_item_width("plot_window") - x_delta
+
+                # Update min widths
+                self.data_pool_min_width = 300
+                self.plot_window_min_width = max(
+                    dpg.get_item_rect_size(self.__timeline.widget_group)[0] + 100,
+                    dpg.get_item_rect_size("plot_buttons")[0] + 100,
+                )
+
+                # Limit new widths
+                if width_right < self.plot_window_min_width:
+                    width_right = self.plot_window_min_width
+                    width_left = self.viewport_size[0] - width_right
+                if width_left < self.data_pool_min_width:
+                    width_left = self.data_pool_min_width
+                    width_right = self.viewport_size[0] - width_left
+
+                # Resize child windows and active plot
+                dpg.configure_item("data_pool_window", width=width_left)
+                dpg.configure_item("plot_window", width=width_right)
+                self.scale[0] = width_left / self.viewport_size[0]
+                self.scale[1] = 1 - self.scale[0]
+                self.resizeCb()
+
+        with dpg.item_handler_registry() as item_handler:
+            dpg.add_item_clicked_handler(callback=clickedCb)
+        dpg.bind_item_handler_registry(item=separator, handler_registry=item_handler)
+
     # -----------------------------------------
     # Plots
     # -----------------------------------------
@@ -299,9 +369,7 @@ class DearBagPlayer:
 
     def createTimeLines(self):
         # Check if act_plot is deleted by users
-        try:
-            dpg.get_item_info(dpg.get_item_user_data(self.tab_bar)['act_plot'])
-        except:
+        if not dpg.does_item_exist(dpg.get_item_user_data(self.tab_bar)['act_plot']):
             return
 
         act_plot = dpg.get_item_user_data(self.tab_bar)['act_plot']
@@ -317,9 +385,7 @@ class DearBagPlayer:
 
     def createTimePoints(self):
         # Check if act_plot is deleted by users
-        try:
-            dpg.get_item_info(dpg.get_item_user_data(self.tab_bar)['act_plot'])
-        except:
+        if not dpg.does_item_exist(dpg.get_item_user_data(self.tab_bar)['act_plot']):
             return
 
         act_plot = dpg.get_item_user_data(self.tab_bar)['act_plot']
@@ -347,6 +413,22 @@ class DearBagPlayer:
             dpg.set_value(item, False)
             dpg.get_item_user_data(self.data_pool_window).remove(item)
 
+    def createErrorPopup(self, error_text, popup_width=300, popup_height=60, min_size=(150, 30)):
+        viewport_pos = dpg.get_viewport_pos()
+        viewport_width = dpg.get_viewport_width()
+        viewport_height = dpg.get_viewport_height()
+        popup_pos = [
+            int(viewport_pos[0] + viewport_width // 2 - popup_width // 2),
+            int(viewport_pos[1] + viewport_height // 2 - popup_height // 2),
+        ]
+        with dpg.mutex():
+            with dpg.window(pos=popup_pos, modal=True, no_close=True, no_move=False,
+                            no_title_bar=True, no_background=False, no_resize=True,
+                            no_scrollbar=True, width=popup_width, height=popup_height,
+                            min_size=min_size) as value_error_popup:
+                dpg.add_text(error_text)
+                dpg.add_button(label="OK", callback=lambda: dpg.delete_item(value_error_popup))
+
     def commonDropCallback(self, yaxis, app_data):
 
         if self.xy_plot_enabled:
@@ -357,8 +439,11 @@ class DearBagPlayer:
 
             # Check they belongs to the same topic and bag
             if datax[3] != datay[3] or datax[4] != datay[4]:
-                # TODO: use popup and return instead of Exception
-                raise ValueError("XY plot must comes from the same bag and topic!")
+                self.createErrorPopup(
+                    "XY plot must comes from the same bag and topic!",
+                    popup_width=350, popup_height=60
+                )
+                return
 
             # Plot line series
             bag_name = os.path.splitext(datax[4])[0]
@@ -382,8 +467,11 @@ class DearBagPlayer:
             datay = dpg.get_item_user_data(app_data[1])
             # Check they belongs to the same topic and bag
             if datax[3] != datay[3] or datax[4] != datay[4]:
-                # TODO: use popup and return instead of Exception
-                raise ValueError("Data vs. s plot must comes from the same bag and topic!")
+                self.createErrorPopup(
+                    "Data vs. s plot must come from the same bag and topic!",
+                    popup_width=400, popup_height=60
+                )
+                return
 
             # Plot line series
             bag_name = os.path.splitext(datax[4])[0]
@@ -429,7 +517,11 @@ class DearBagPlayer:
     def deleteAllSeriesCb(self, sender, app_data, user_data):
         """user_data stores the tag of selected series"""
         plot = dpg.get_item_parent(user_data)
-        dpg.delete_item(plot, children_only=True)
+        series_list = dpg.get_item_children(plot, 1)
+        for series in series_list:
+            if (self.vlines is None or series not in self.vlines) and \
+               (self.xypoints is None or series not in self.xypoints):
+                dpg.delete_item(series)
 
     def plotDropCallback(self, sender, app_data, user_data):
         yaxis = dpg.get_item_info(sender)["children"][1][1]
@@ -531,6 +623,7 @@ class DearBagPlayer:
             dpg.add_input_text(label="", hint="<new name>", on_enter=True,
                                user_data=(user_data[0], rename_win),
                                callback=self.renameWindowCb)
+            dpg.focus_item(dpg.last_item())
 
     def renameWindowCb(self, sender, app_data, user_data):
         """
@@ -588,6 +681,7 @@ class DearBagPlayer:
         self.clearTimeLinesAndPoints()
         dpg.get_item_user_data(sender)['act_tab'] = app_data
         dpg.get_item_user_data(sender)['act_plot'] = dpg.get_item_children(app_data)[1][0]
+        self.resizeCb()
 
     def deleteClosedTab(self):
         for tab in dpg.get_item_children(self.tab_bar)[1]:
@@ -616,9 +710,19 @@ class DearBagPlayer:
     # File Import
     # -----------------------------------------
 
-    def selectDataFiles(self, sender, app_data, user_data):
-        print("Sender: ", sender)
-        print("App Data: ", app_data)
+    def selectDataFilesCb(self, sender, app_data, user_data):
+        """
+        :param sender: file_dialog tag
+        :param app_data:
+            {'file_path_name': ...,
+             'file_name': ...,
+             'current_path': ...,
+             'current_filter': ...,
+             'min_size': ..., 'max_size': ...,
+             'selections': {'<file_name>': '<file_path>', ...},
+             }
+        :param user_data: None
+        """
         for key, value in app_data["selections"].items():
             self.bag_files.append(value)
             self.bag_files_name.append(key)
@@ -664,50 +768,60 @@ class DearBagPlayer:
     # Main Entry
     # -----------------------------------------
 
-    def eventHandler(self, sender, data):
-        # KeyDown data = [key, seconds]; KeyRelease data = key
+    def specialPlotKeyEventCb(self, sender, data):
+        """
+        Key event callback for special plot
+
+        Get event type with: `event_type = dpg.get_item_info(sender)["type"]`
+        - "mvAppItemType::mvKeyPressHandler"
+        - "mvAppItemType::mvKeyReleaseHandler"
+        - "mvAppItemType::mvKeyDownHandler" (much more frequently)
+
+        :param sender: handler tag
+        :param data: KeyPress/KeyRelease data - key, KeyDown data: [key, elapsed_time]
+        """
         event_type = dpg.get_item_info(sender)["type"]
-        # print(data, type(data))
-        if event_type == "mvAppItemType::mvKeyDownHandler" and data[0] == dpg.mvKey_Control:
-            self.xy_plot_enabled = True
-            # print("XY Pos Enabled!")
-        elif event_type == "mvAppItemType::mvKeyReleaseHandler" and data == dpg.mvKey_Control:
-            self.xy_plot_enabled = False
-            # print("XY Pos Disabled!")
-        if event_type == "mvAppItemType::mvKeyDownHandler" and data[0] == dpg.mvKey_Shift:
-            self.s_length_plot_enabled = True
-            # print("S length plot Enabled!")
-        elif event_type == "mvAppItemType::mvKeyReleaseHandler" and data == dpg.mvKey_Shift:
-            self.s_length_plot_enabled = False
-            # print("S length plot Disabled!")
+        if data == dpg.mvKey_Control:
+            self.xy_plot_enabled = True if event_type == "mvAppItemType::mvKeyPressHandler" else False
+        elif data == dpg.mvKey_Shift:
+            self.s_length_plot_enabled = True if event_type == "mvAppItemType::mvKeyPressHandler" else False
 
     def run(self):
         # Call this function at the beginning in every DearPyGui application
         dpg.create_context()
 
-        # Viewport
-        dpg.create_viewport(title="DearBagPlayer", width=1500, height=900, x_pos=0, y_pos=0)
-
         # Icon TODO
         # dpg.set_viewport_small_icon("path/to/icon.ico")
         # dpg.set_viewport_large_icon("path/to/icon.ico")
 
-        # Initialization
-        dpg.setup_dearpygui()
-        dpg.show_viewport()
+        # Viewport
+        dpg.create_viewport(title="DearBagPlayer", width=800, height=600, x_pos=0, y_pos=0,
+                            min_width=800, min_height=600)
+
+        self.viewport_size[0] = dpg.get_viewport_width()
+        self.viewport_size[1] = dpg.get_viewport_height()
+
+        # Plot Handlers
+        with dpg.handler_registry(tag="special_plot_key_event_handler"):  # show=True by default
+            dpg.add_key_release_handler(key=dpg.mvKey_Control)
+            dpg.add_key_press_handler(key=dpg.mvKey_Control)
+            dpg.add_key_release_handler(key=dpg.mvKey_Shift)
+            dpg.add_key_press_handler(key=dpg.mvKey_Shift)
+
+        for handler in dpg.get_item_children("special_plot_key_event_handler", 1):
+            dpg.set_item_callback(handler, self.specialPlotKeyEventCb)
+
+        with dpg.item_handler_registry(tag="tab_clicked_handler"):
+            dpg.add_item_clicked_handler(button=1, callback=self.tabClickedMenuCb)
 
         # Viewport menu bar
         with dpg.file_dialog(directory_selector=False, show=False, file_count=10,
-                             width=600, height=600,
-                             callback=self.selectDataFiles) as file_dialog_tag:
-            dpg.add_file_extension(".*")
-            dpg.add_file_extension("", color=(150, 255, 150, 255))
-            dpg.add_file_extension("Source files (*.cpp *.h *.hpp){.cpp,.h,.hpp}", color=(0, 255, 255, 255))
-            dpg.add_file_extension(".h", color=(255, 0, 255, 255), custom_text="[header]")
-            dpg.add_file_extension(".py", color=(0, 255, 0, 255), custom_text="[Python]")
+                             width=600, height=400, modal=True,
+                             callback=self.selectDataFilesCb) as file_dialog_tag:
             dpg.add_file_extension(".bag", color=(0, 255, 0, 255), custom_text="[rosbag]")
+            dpg.add_file_extension(".*")
 
-        with dpg.viewport_menu_bar():
+        with dpg.viewport_menu_bar(tag="menubar"):
             with dpg.menu(label="Files"):
                 dpg.add_menu_item(label="Import Data", callback=lambda: dpg.show_item(file_dialog_tag))
 
@@ -720,46 +834,51 @@ class DearBagPlayer:
                 dpg.add_menu_item(label="Show Font Manager", callback=lambda: dpg.show_tool(dpg.mvTool_Font))
                 dpg.add_menu_item(label="Show Item Registry", callback=lambda: dpg.show_tool(dpg.mvTool_ItemRegistry))
 
-        # Data series list
-        self.data_pool_window = dpg.add_window(label="Data Pool", height=650, width=400, user_data=list())
+        # Primary Window
+        dpg.add_window(tag="primary_window")
+        dpg.set_primary_window("primary_window", True)
+        dpg.configure_item("primary_window", no_scrollbar=True)
 
-        # Handlers
-        with dpg.handler_registry(tag="__demo_keyboard_handler"):  # show=True by default
-            dpg.add_key_down_handler(key=dpg.mvKey_Control)
-            dpg.add_key_release_handler(key=dpg.mvKey_Control)
-            dpg.add_key_press_handler(key=dpg.mvKey_Control)
-            dpg.add_key_down_handler(key=dpg.mvKey_Shift)
-            dpg.add_key_release_handler(key=dpg.mvKey_Shift)
-            dpg.add_key_press_handler(key=dpg.mvKey_Shift)
+        # Make top borders of child windows visible
+        dpg.add_child_window(parent="primary_window", width=10, height=10)
 
-        for handler in dpg.get_item_children("__demo_keyboard_handler", 1):
-            dpg.set_item_callback(handler, self.eventHandler)
+        # Workspace
+        with dpg.group(horizontal=True, parent="primary_window"):
+            self.data_pool_window = dpg.add_child_window(
+                label="Data Pool", user_data=list(),
+                tag="data_pool_window", width=int(self.scale[0]*self.viewport_size[0]),
+            )
 
-        with dpg.item_handler_registry(tag="tab_clicked_handler"):
-            dpg.add_item_clicked_handler(callback=self.tabClickedMenuCb)
+            self.addVerticalSeparator()
 
-        # Plot window
-        with dpg.window(label="Plot Window", pos=(420, 0), height=800, width=810):
-            with dpg.group(horizontal=True):
-                dpg.add_button(label="Split Horizontally", callback=self.splitHorizontallyCb)
-                dpg.add_button(label="Split Vertically", callback=self.splitVerticallyCb)
-                dpg.add_button(label="Clear", callback=self.clearCb)
+            with dpg.child_window(label="Plot Window", tag="plot_window",
+                                  width=int(self.scale[1]*self.viewport_size[0])):
+                with dpg.group(horizontal=True, tag="plot_buttons"):
+                    dpg.add_button(label="Split Horizontally", callback=self.splitHorizontallyCb)
+                    dpg.add_button(label="Split Vertically", callback=self.splitVerticallyCb)
+                    dpg.add_button(label="Clear", callback=self.clearCb)
 
-            with dpg.tab_bar(user_data={"act_tab": None, "act_plot": None, "plot_pages": 1},
-                             reorderable=True, callback=self.updateActCb) as self.tab_bar:
-                with dpg.tab(label=f"Plot {dpg.get_item_user_data(self.tab_bar)['plot_pages']}",
-                             closable=True) as tab_tag:
-                    dpg.bind_item_handler_registry(tab_tag, "tab_clicked_handler")
-                    dpg.get_item_user_data(self.tab_bar)['act_tab'] = tab_tag
-                    with dpg.subplots(rows=1, columns=1, no_title=True, height=600, width=800):
-                        dpg.get_item_user_data(self.tab_bar)['act_plot'] = dpg.last_item()
-                        self.addPlot()
-                dpg.add_tab_button(label="+", tag="Add Plot Button", callback=self.addPlotPageCb, trailing=True)
+                with dpg.tab_bar(user_data={"act_tab": None, "act_plot": None, "plot_pages": 1},
+                                 reorderable=True, callback=self.updateActCb) as self.tab_bar:
+                    with dpg.tab(label=f"Plot {dpg.get_item_user_data(self.tab_bar)['plot_pages']}",
+                                 closable=True) as tab_tag:
+                        dpg.bind_item_handler_registry(tab_tag, "tab_clicked_handler")
+                        dpg.get_item_user_data(self.tab_bar)['act_tab'] = tab_tag
+                        with dpg.subplots(rows=1, columns=1, no_title=True):
+                            dpg.get_item_user_data(self.tab_bar)['act_plot'] = dpg.last_item()
+                            self.addPlot()
+                    dpg.add_tab_button(label="+", tag="Add Plot Button", callback=self.addPlotPageCb, trailing=True)
 
-            self.__timeline.createWidgets()
+                self.__timeline.createWidgets()
 
-        # The Primary Window
-        # dpg.set_primary_window("Primary Window", True)
+        # Bind resize handler
+        with dpg.item_handler_registry(tag="resize_handler"):
+            dpg.add_item_resize_handler(callback=self.resizeCb)
+
+        dpg.bind_item_handler_registry("primary_window", "resize_handler")
+
+        dpg.setup_dearpygui()
+        dpg.show_viewport()
 
         # Start DPG application
         # dpg.start_dearpygui()
